from jsengine.util import to_unicode, json_encoder


# Some simple compatibility processing
init_global_script = u'''\
if (typeof global === 'undefined')
    if (typeof Proxy === 'function')
        global = new Proxy(this, {})
    else
        global = this
if (typeof globalThis === 'undefined')
    globalThis = this
'''

init_del_gobject_script = u'''\
if (typeof {gobject} !== 'undefined')
    delete {gobject}
'''


class AbstractJSEngine(object):

    def __init__(self, source=u'', init_global=False, init_del_gobjects=[]):
        '''Create a JSEngine content.

        Params:
            source:
                any valid Javescript.
            init_global:
                set True to ensure `global` and `globalThis` are available.
            init_del_gobjects:
                use to delete some variables in the global.
        '''
        self._source = []
        init_script = []
        if init_global:
            init_script.append(init_global_script)
        if init_del_gobjects:
            for gobject in init_del_gobjects:
                init_script.append(init_del_gobject_script.format(gobject=gobject))
        self.append(u''.join(init_script))
        self.append(source)

    @property
    def source(self):
        '''All the inputted Javascript code.'''
        return u'\n'.join(self._source)

    def _append_source(self, code):
        if code:
            self._source.append(code)

    def _check_code(self, code):
        # Input unicode
        code = to_unicode(code)
        first_c = code.lstrip()[:1]
        if first_c:
            # Simple separator check
            last_c = self._source and self._source[-1].rstrip()[-1:] or ''
            if last_c and (first_c in u'([`/' and last_c.isdecimal() or
                           first_c in u'`/' and last_c not in ',;}+-*/%!=<>&|'):
                code = u';' + code
            return code

    def append(self, code):
        '''Run Javascript code and return none.'''
        code = self._check_code(code)
        if code:
            self._append(code)

    def eval(self, code):
        '''Run Javascript code and return result.'''
        code = self._check_code(code)
        if code:
            return self._eval(code)

    def call(self, identifier, *args):
        '''Use name string and arguments to call Javascript function.'''
        chunks = json_encoder.iterencode(args, _one_shot=True)
        chunks = [to_unicode(chunk) for chunk in chunks]
        args = u''.join(chunks)[1:-1]
        code = u'{identifier}({args})'.format(**vars())
        return self._eval(code)

    def _append(self, code):
        raise NotImplementedError('Method must be implemented by subclass')

    def _eval(self, code):
        raise NotImplementedError('Method must be implemented by subclass')



from jsengine.abstract import AbstractJSEngine
from jsengine.exceptions import *
import jsengine.detect as _d
import json


class InternalJSEngine(AbstractJSEngine):
    '''Wrappered for Internal(DLL) Javascript interpreter.'''

    def __init__(self, *args, **kwargs):
        self._context = self.Context(self)
        AbstractJSEngine.__init__(self, *args, **kwargs)

    def _append(self, code):
        self._context.eval(code, eval=False, raw=True)

    def _eval(self, code):
        return self._context.eval(code)

    class Context:
        def __init__(self, engine):
            raise NotImplementedError('Class `Context` must be implemented by subclass')


class ChakraJSEngine(InternalJSEngine):
    '''Wrappered for system's built-in Chakra or PyChakra(ChakraCore).'''

    def __init__(self, source=u'', init_global=False, init_del_gobjects=[]):
        if not _d.chakra_available:
            msg = 'No supported Chakra binary found on your system!'
            if _d.quickjs_available:
                msg += ' Please install PyChakra or use QuickJSEngine.'
            elif _d.external_interpreter:
                msg += ' Please install PyChakra or use ExternalJSEngine.'
            else:
                msg += ' Please install PyChakra.'
            raise RuntimeError(msg)
        InternalJSEngine.__init__(self, source, init_global, init_del_gobjects)

    __init__.__doc__ = AbstractJSEngine.__init__.__doc__

    class Context:
        def __init__(self, engine):
            self._engine = engine
            self._context = _d.ChakraHandle()

        def eval(self, code, eval=True, raw=False):
            self._engine._append_source(code)
            ok, result = self._context.eval(code, raw=raw)
            if ok:
                if eval:
                    return result
            else:
                raise ProgramError(str(result))


class QuickJSEngine(InternalJSEngine):
    '''Wrappered for QuickJS python binding quickjs.'''

    def __init__(self, source=u'', init_global=False, init_del_gobjects=[]):
        if not _d.quickjs_available:
            msg = 'No supported QuickJS package found on custom python environment!'
            if _d.chakra_available:
                msg += ' Please install python package quickjs or use ChakraJSEngine.'
            elif _d.external_interpreter:
                msg += ' Please install python package quickjs or use ExternalJSEngine.'
            else:
                msg += ' Please install python package quickjs.'
            raise RuntimeError(msg)
        InternalJSEngine.__init__(self, source, init_global, init_del_gobjects)

    __init__.__doc__ = AbstractJSEngine.__init__.__doc__

    class Context:
        def __init__(self, engine):
            self._engine = engine
            self._context = _d.quickjs.Context()
            self.typeof = self.Function(self, self._context.eval(u'(obj => typeof obj)'))

        def eval(self, code, eval=True, raw=False):
            self._engine._append_source(code)
            try:
                result = self._context.eval(code)
            except _d.quickjs.JSException as e:
                raise ProgramError(*e.args)
            else:
                if eval:
                    if raw or not isinstance(result, _d.quickjs.Object):
                        return result
                    elif callable(result) and self.typeof(result) == u'function':
                        return self.Function(self, result)
                    else:
                        return json.loads(result.json())

        class Function:
            # PetterS/quickjs/Issue7
            # Escape StackOverflow when calling function outside
            def __init__(self, context, function):
                self._context = context
                self._function = function

            def __call__(self, *args):
                return self._function(*args)



'''This is a Python binding to Microsoft Chakra Javascript engine.
Forked from PyChakra (https://github.com/zhengrenzhe/PyChakra) to support
Windows' built-in Chakra.
'''


import ctypes as _ctypes
import json


# load Windows' built-in chakra binary
try:
    chakra = _ctypes.windll.Chakra
    chakra_available = True
except:
    chakra_available = False


class ChakraHandle():

    def __init__(self):
        # create chakra runtime and context
        runtime = _ctypes.c_void_p()
        chakra.JsCreateRuntime(0, 0, point(runtime))

        context = _ctypes.c_void_p()
        chakra.JsCreateContext(runtime, point(context))
        chakra.JsSetCurrentContext(context)

        self.__runtime = runtime
        self.__context = context
        self.__chakra = chakra

        # get JSON.stringify reference, and create its called arguments array
        stringify = self.eval('JSON.stringify', raw=True)[1]
        undefined = _ctypes.c_void_p()
        chakra.JsGetUndefinedValue(point(undefined))
        args = (_ctypes.c_void_p * 2)()
        args[0] = undefined

        self.__jsonStringify = stringify
        self.__jsonStringifyArgs = args

    def __del__(self):
        self.__chakra.JsDisposeRuntime(self.__runtime)

    def eval(self, script, raw=False):
        '''Eval javascript string

        Examples:
            .eval('(()=>2)()') // (True, 2)
            .eval('(()=>a)()') // (False, "ReferenceError: 'a' is not defined")

        Parameters:
            script(str): javascript code string
            raw(bool?): whether return result as chakra JsValueRef directly
                        (optional, default is False)

        Returns:
            (bool, result)
            bool: indicates whether javascript is running successfully.
            result: if bool is True, result is the javascript running
                        return value.
                    if bool is False and result is string, result is the
                        javascript running exception
                    if bool is False and result is number, result is the
                        chakra internal error code
        '''

        # TODO: may need a thread lock, if running multithreading
        self.__chakra.JsSetCurrentContext(self.__context)

        js_source = _ctypes.c_wchar_p('')
        js_script = _ctypes.c_wchar_p(script)

        result = _ctypes.c_void_p()
        err = self.__chakra.JsRunScript(js_script, 0, js_source, point(result))

        # no error
        if err == 0:
            if raw:
                return True, result
            else:
                return self.__js_value_to_py_value(result)

        return self.__get_error(err)

    def __js_value_to_py_value(self, js_value):
        args = self.__jsonStringifyArgs
        args[1] = js_value

        # value => json
        result = _ctypes.c_void_p()
        err = self.__chakra.JsCallFunction(
            self.__jsonStringify, point(args), 2, point(result))

        if err == 0:
            result = self.__js_value_to_str(result)
            if result == 'undefined':
                result = None
            else:
                # json => value
                result = json.loads(result)
            return True, result

        return self.__get_error(err)

    def __get_error(self, err):
        # js exception or other error
        if err == 196609:
            err = self.__get_exception()
        return False, err

    def __get_exception(self):
        exception = _ctypes.c_void_p()
        self.__chakra.JsGetAndClearException(point(exception))
        return self.__js_value_to_str(exception)

    def __js_value_to_str(self, js_value):
        js_value_ref = _ctypes.c_void_p()
        self.__chakra.JsConvertValueToString(js_value, point(js_value_ref))

        str_p = _ctypes.c_wchar_p()
        str_l = _ctypes.c_size_t()
        self.__chakra.JsStringToPointer(js_value_ref, point(str_p), point(str_l))
        return str_p.value


def point(any):
    return _ctypes.byref(any)




from __future__ import print_function

import os
import sys
import platform
from jsengine.util import which


chakra_available = False
quickjs_available = False
external_interpreter = None

# PyChakra
try:
    from PyChakra import Runtime as ChakraHandle, get_lib_path
    if not os.path.isfile(get_lib_path()):
        raise RuntimeError
except (ImportError, RuntimeError):
    pass
else:
    chakra_available = True

# PyQuickJS
try:
    import quickjs
except ImportError:
    pass
else:
    quickjs_available = True

# macOS: built-in JavaScriptCore
if platform.system() == 'Darwin':
    # jsc lives on a new path since macOS Catalina
    jsc_paths = ['/System/Library/Frameworks/JavaScriptCore.framework/Versions/A/Resources/jsc',
                 '/System/Library/Frameworks/JavaScriptCore.framework/Versions/A/Helpers/jsc']
    for interpreter in jsc_paths:
        external_interpreter = which(interpreter)
        if external_interpreter:
            break

# Windows: built-in Chakra, or Node.js, QuickJS if installed
elif platform.system() == 'Windows':
    if not chakra_available:
        from jsengine.chakra_win import ChakraHandle, chakra_available

    for interpreter in ('qjs', 'node', 'nodejs'):
        external_interpreter = which(interpreter)
        if external_interpreter:
            break

    if not chakra_available and not quickjs_available and external_interpreter is None:
        print('Please install PyChakra or Node.js!', file=sys.stderr)

# Linux: Gjs on Gnome, CJS on Cinnamon, or JavaScriptCore, Node.js if installed
else:
    for interpreter in ('gjs', 'cjs', 'jsc', 'qjs', 'nodejs', 'node'):
        external_interpreter = which(interpreter)
        if external_interpreter:
            break

    if not chakra_available and not quickjs_available and external_interpreter is None:
        if platform.system() == 'Linux':
            print('''\
Please install at least one of the following Javascript interpreter.
python packages: PyChakra, quickjs
applications: Gjs, CJS, QuickJS, JavaScriptCore, Node.js.''', file=sys.stderr)

        else:
            print('''\
Sorry, JSEngine is currently not supported officially on your system.
Please try install one of the following Javascript interpreter.
applications: Gjs, CJS, QuickJS, JavaScriptCore, Node.js.''', file=sys.stderr)



import sys
import json

try:
    from shutil import which
except ImportError:
    from distutils.spawn import find_executable as which


if sys.version_info > (3,):
    unicode = str

def to_unicode(s):
    if not isinstance(s, unicode):
        s = s.decode('utf8')
    return s

def to_bytes(s):
    if isinstance(s, unicode):
        s = s.encode('utf8')
    return s

def json_encoder_fallback(o):
    # Allow bytes (python3)
    if isinstance(o, bytes):
        return to_unicode(o)
    return json.JSONEncoder.default(json_encoder, o)

json_encoder = json.JSONEncoder(
    skipkeys=True,
    ensure_ascii=False,
    check_circular=True,
    allow_nan=True,
    indent=None,
    separators=None,
    default=json_encoder_fallback,
)

json_encoder_ensure_ascii = json.JSONEncoder(
    skipkeys=True,
    ensure_ascii=True,
    check_circular=True,
    allow_nan=True,
    indent=None,
    separators=None,
    default=None,
)



from __future__ import print_function

from jsengine.abstract import AbstractJSEngine
from jsengine.exceptions import *
from jsengine.util import which, to_bytes, json_encoder, json_encoder_ensure_ascii
from subprocess import Popen, PIPE, list2cmdline
import jsengine.detect as _d
import tempfile
import os
import sys
import json

# The maximum length of command string
if os.name == 'posix':
    # Used in Unix is ARG_MAX in conf
    ARG_MAX = int(os.popen('getconf ARG_MAX').read())
else:
    # Used in Windows CreateProcess is 32K
    ARG_MAX = 32 * 1024


class ExternalJSEngine(AbstractJSEngine):
    '''Wrappered for external Javascript interpreter.'''

    def __init__(self, source=u'', init_global=False, init_del_gobjects=[],
                       interpreter=None, **kwargs):
        '''
            (interpreter, **kwargs):
                same as ExternalInterpreter.__init__
        '''
        if isinstance(interpreter, str):
            interpreter = ExternalInterpreter.get(interpreter, **kwargs)
        if isinstance(interpreter, ExternalInterpreter):
            self.interpreter = interpreter
        elif isinstance(_d.external_interpreter, ExternalInterpreter):
            self.interpreter = _d.external_interpreter
        else:
            msg = 'No supported external Javascript interpreter found on your system!'
            if _d.chakra_available:
                msg += ' Please install one or use ChakraJSEngine.'
            elif _d.quickjs_available:
                msg += ' Please install one or use QuickJSEngine.'
            else:
                msg += ' Please install one.'
            raise RuntimeError(msg)
        # Del 'exports' to ignore import error, e.g. Node.js
        init_del_gobjects = list(init_del_gobjects) + ['exports']
        AbstractJSEngine.__init__(self, source, init_global, init_del_gobjects)

    __init__.__doc__ = AbstractJSEngine.__init__.__doc__ + __init__.__doc__[9:]

    def _append(self, code):
        self._append_source(code)

    def _eval(self, code):
        self._append_source(code)
        code = self._inject_script()

        evalstring = False
        if self.interpreter.evalstring:
            try:
                output = self._run_interpreter_with_string(code)
                evalstring = True
            except ValueError:
                pass
            except RuntimeError:
                self.interpreter.evalstring = False

        if not evalstring and not self.interpreter.tempfile:
            try:
                output = self._run_interpreter_with_pipe(code)
            except RuntimeError:
                self.interpreter.tempfile = True

        while True:
            if not evalstring and self.interpreter.tempfile:
                output = self._run_interpreter_with_tempfile(code)

            output = output.replace(u'\r\n', u'\n').replace(u'\r', u'\n')
            # Search result in the last 5 lines of output
            for result_line in output.split(u'\n')[-5:]:
                if result_line[:9] == u'["result"':
                    break
            try:
                _, ok, result = json.loads(result_line)
            except json.decoder.JSONDecodeError as e:
                if not evalstring and self.interpreter.tempfile:
                    raise RuntimeError('%s:\n%s' % (e, output))
                else:
                    evalstring = False
                    self.interpreter.tempfile = True
                    continue
            if ok:
                return result
            else:
                raise ProgramError(result)

    def _run_interpreter(self, cmd, input=None):
        stdin = PIPE if input else None
        p = Popen(cmd, stdin=stdin, stdout=PIPE, stderr=PIPE)
        stdout_data, stderr_data = p.communicate(input=input)
        if p.returncode != 0:
            raise RuntimeError('%r returns non-zero value! Error msg: %s' %
                               (_d.external_interpreter, stderr_data.decode('utf8')))
        elif stderr_data:
            print("%r has warnings:" % _d.external_interpreter, stderr_data.decode('utf8'))
        # Output unicode
        return stdout_data.decode('utf8')

    def _run_interpreter_with_string(self, code):
        cmd = self.interpreter.command + [self.interpreter.evalstring, code]
        if len(list2cmdline(cmd)) > ARG_MAX:  # Direct compare, don't wait an Exception
            raise ValueError('code length is too long to run as a command')
        return self._run_interpreter(cmd)

    def _run_interpreter_with_pipe(self, code):
        # Input bytes
        return self._run_interpreter(self.interpreter.command, input=to_bytes(code))

    def _run_interpreter_with_tempfile(self, code):
        fd, filename = tempfile.mkstemp(prefix='execjs', suffix='.js')
        try:
            # Write bytes
            with open(fd, 'wb') as fp:
                fp.write(to_bytes(code))
            return self._run_interpreter(self.interpreter.command + [filename])
        finally:
            os.remove(filename)

    def _inject_script(self):
        if self.interpreter.evalstring:
            source = json_encoder_ensure_ascii.encode(self.source)
        else:
            source = json_encoder.encode(self.source)
        return injected_script.format(source=source)


class ExternalInterpreter:
    '''External interpreter setting.'''

    @classmethod
    def get(cls, *args, **kwargs):
        '''Same as cls.__init__, the fallback is None.'''
        try:
            return cls(*args, **kwargs)
        except Exception as e:
            print(e, file=sys.stderr)

    def __init__(self, interpreter, name=None, tempfile=False, evalstring=False, args=None):
        '''Create an external interpreter setting.

        params:
            interpreter:
                None means default interpreter by detected,
                or the filename/filepath of a interpreter,
                or a ExternalInterpreter instance.
            name:
                use to get default setting, see ExternalInterpreterNameAlias.
            tempfile:
                whether to use tempfile or not to pass Javascript code.
            evalstring:
                whether to use shell cmd or not to pass Javascript code.
                if set True, default param is `-c`.
                here also accept other valid params likes `--execute`
            args:
                any valid params of the interpreter.
        '''
        path = which(interpreter)
        if path is None:
            raise ValueError('Can not find the given interpreter: %r' % interpreter)
        filename = os.path.basename(path).rsplit('.', 1)[0]
        if name is None:
            name = filename
        name = ExternalInterpreterNameAlias.get(name.lower().replace('.', ''), name)
        if name in DefaultExternalInterpreterOptions:
            tempfile, evalstring, _args = DefaultExternalInterpreterOptions[name]
            if not args:
                args = _args
        self.name = name
        self.path = path
        self.tempfile = tempfile
        # `-e`, `-eval` means run command string as Javascript
        # But some interpreters don't use `-eval`
        if isinstance(evalstring, str) and evalstring.startswith('-'):
            self.evalstring = evalstring
        else:
            self.evalstring = evalstring and '-e' or False
        self.command = [path]
        if args:
            self.command += list(args)

    def __repr__(self):
        return '<ExternalInterpreter %s @ %r>' % (self.name, self.path)


DefaultExternalInterpreterOptions = {
                # tempfile, evalstring, args
    'ChakraCore': [ True, False, []],
       'Node.js': [ True,  True, []],
       'QuickJS': [ True,  True, []],
            'V8': [ True,  True, []],
            'XS': [ True,  True, []],
}

ExternalInterpreterNameAlias = {
    # *1 Unceremonious name is not recommended to be used as the binary name
        'chakracore': 'ChakraCore',
            'chakra': 'ChakraCore',
                'ch': 'ChakraCore',    # *1
               'cjs': 'CJS',
               'gjs': 'Gjs',
    'javascriptcore': 'JavaScriptCore',
               'jsc': 'JavaScriptCore',
            'nodejs': 'Node.js',
              'node': 'Node.js',
           'quickjs': 'QuickJS',
               'qjs': 'QuickJS',
              'qjsc': 'QuickJS',
           'jsshell': 'SpiderMonkey',
      'spidermonkey': 'SpiderMonkey',
                'sm': 'SpiderMonkey',  # *1
                'js': 'SpiderMonkey',  # *1
                'v8': 'V8',            # *1
                'd8': 'V8',            # *1
                'xs': 'XS',            # *1
               'xst': 'XS',
    # Don't use these interpreters
    # They are not compatible with the most used ES6 features
           'duktape': 'Duktape(incompatible)',
               'duk': 'Duktape(incompatible)',
            'hermes': 'Hermes(incompatible)',
           'cscript': 'JScript(incompatible)',
         'phantomjs': 'PhantomJS(incompatible)',
}

# Inject to the script to let it return jsonlized value to python
# Fixed our helper objects
injected_script = u'''\
Object.defineProperty((typeof global !== 'undefined') && global ||
                      (typeof globalThis !== 'undefined') && globalThis ||
                      this, '_JSEngineHelper', {{
    value: {{}},
    writable: false,
    configurable: false
}})
Object.defineProperty(_JSEngineHelper, 'print', {{
    value: function(s) {{
        if (typeof console !== 'undefined' && typeof console.log !== 'undefined')
            console.log(s)
        else if (typeof print === 'undefined')
            print(s)
    }},
    writable: false,
    configurable: false
}})
Object.defineProperty(_JSEngineHelper, 'jsonStringify', {{
    value: JSON.stringify,
    writable: false,
    configurable: false
}})
Object.defineProperty(_JSEngineHelper, 'result', {{
    value: null,
    writable: true,
    configurable: false
}})
Object.defineProperty(_JSEngineHelper, 'status', {{
    value: false,
    writable: true,
    configurable: false
}})
try {{
    _JSEngineHelper.result = eval({source}), _JSEngineHelper.status = true
}}
catch (err) {{
    _JSEngineHelper.result = err.toString(), _JSEngineHelper.status = false
}}
try {{
    _JSEngineHelper.print('\\n' + _JSEngineHelper.jsonStringify(
        ["result", _JSEngineHelper.status, _JSEngineHelper.result]))
}}
catch (err) {{
    _JSEngineHelper.print(
        '\\n["result", false, "Script returns a value with an unsupported type"]')
}}
'''



#!/usr/bin/env python

import os
import re

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def read_file(*paths):
    with open(os.path.join(here, *paths), 'r') as fp:
        return fp.read()

def get_version():
    content = read_file(package_name, '__init__.py')
    version_match = re.search('^__version__ = [\'"]([^\'"]+)', content, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')

here = os.path.abspath(os.path.dirname(__file__))
package_name = 'jsengine'

setup(
    name='jsengine',
    version=get_version(),
    author='SeaHOH',
    author_email='seahoh@gmail.com',
    url='https://github.com/SeaHOH/jsengine',
    license='MIT',
    description=('JSEngine is a simple wrapper of Javascript engines.'),
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    keywords='javascript js engine node chakra quickjs execjs',
    packages=[package_name],
    zip_safe=True,
    platforms='any',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: JavaScript',
        'Topic :: Software Development :: Interpreters',
        'Topic :: Utilities'
    ],
    python_requires='>=2.7',
)



# Copyright (c) 2019 coslyk
# Copyright (c) 2019 - 2021 SeaHOH <SeaHOH@gmail.com>


import platform
import jsengine.detect as _d
from jsengine.exceptions import *
from jsengine.internal import QuickJSEngine, ChakraJSEngine
from jsengine.external import ExternalJSEngine, ExternalInterpreter


__version__ = '1.0.1'

__all__ = ['JSEngine', 'ChakraJSEngine', 'QuickJSEngine', 'ExternalJSEngine',
           'ExternalInterpreter', 'set_external_interpreter',
           'Error', 'RuntimeError', 'ProgramError',
           'jsengine', 'eval']


def set_external_interpreter(interpreter, *args, **kwargs):
    '''
    Set default an external interpreter, return the result status.
    Same arguments as the ExternalInterpreter.
    '''
    interpreter = ExternalInterpreter.get(interpreter, *args, **kwargs)
    if interpreter:
        _d.external_interpreter = interpreter
    return interpreter


if _d.external_interpreter:
    _d.external_interpreter = ExternalInterpreter(_d.external_interpreter)

# Prefer InternalJSEngine (via dynamic library loading)
if _d.quickjs_available:
    JSEngine = QuickJSEngine
elif _d.chakra_available:
    JSEngine = ChakraJSEngine
elif _d.external_interpreter:
    JSEngine = ExternalJSEngine
else:
    JSEngine = None


def jsengine():
    '''Create a context of the default Javascript engine.'''
    if JSEngine:
        return JSEngine()

    if platform.system() in ('Darwin', 'Windows', 'Linux'):
        msg = 'No supported Javascript interpreter has been found'
    else:
        msg = 'Your system does not be supported officially'
    msg += ', please try install one of Gjs, CJS, QuickJS, JavaScriptCore, Node.js'
    raise RuntimeError(msg)

def eval(source):
    '''Run Javascript code use the default engine and return result.'''
    return jsengine().eval(source)

